from .utils import getenv, save_object

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class Data:

    data = {}

    def set_data(self, name: str, value: list):
        if name == '':
            raise Exception(f'Попытка записать в переменную памяти с именем ""')
        self.data[name.lower()] = []
        self.data[name.lower()].extend(value)
        # Залогируем выполненныю задачу
        if getenv("SAVE_OBJECTS", "0") == "1":        
            save_object(value_dict = value, filename = name)

    def get_data(self, name: str) -> list:
        if name.lower() in self.data:
            return self.data[name.lower()]
        else:
            raise Exception(f'Не определена переменная {name}')
        
    def get_data_keys(self) -> list:
        return [key for key in self.data]    

mems = Data()


import re
# Стандартное логирование
import logging

from .data import mems

# Установить текущи логгер
logger = logging.getLogger(__name__)

def list2dict(params_as_list: list) -> dict:
    __dict = {}
    for elem in params_as_list:
        if isinstance(elem, dict):
            for key, value in elem.items():
                __dict[key] = value
    return __dict        

class Param:

    def prepare(self, value: dict, key: str = None) -> dict:
        """Скрыть пароли и токены в словаре

        :param dict value: исходный словарь
        :param str key: Ключ
        :return dict: преобразованный словарь
        """
        if value is None:
            return None
        elif isinstance(value, dict):
            return {k: self.prepare(v, k) for k, v in value.items()}
        elif isinstance(value, list):
            return [self.prepare(elem) for elem in value]
        else:
            if isinstance(value, str):
                m = re.match('^~(.+)$', value.strip())
                if m:
                    # logger.info(f'GET DATA {m.group(1)=}')
                    return mems.get_data(m.group(1))
            return (value)        

    def __init__(self, param:dict = None):
        """Инициализировать класс параметра

        :param object pgm: Ссылка на объект программы Pgm
        :param int stage_id: идентификатор stage
        :param str name: наименование параметра
        :param list tps: список доступных типов
        """
        if param is not None:
            # Рекурсивно разобрать косвенную адресацию
            self.param = self.prepare(value = param)

    def as_list(self):
        if self.param is None:
            return []
        else:    
            if isinstance(self.param, list):
                result = []
                for elem in self.param:
                    if isinstance(elem, list):
                        result.extend(elem)
                    else:
                        result.append(elem)
                return result        
            else:
                return [self.param]    

    def as_list2(self):
        return self.as_list()

    def as_string(self):
        return "\n".join([f'{s}' for s in self.as_list2()])

    def as_dict(self):
        return list2dict(self.as_list())

    def as_int(self):
        if len(self.as_list2()) == 1:
            try:
                return int(self.as_list2()[0])
            except:
                return 0
        else:
            return 0

    def as_float(self):
        if len(self.as_list2()) == 1:
            try:
                return float(self.as_list2()[0])
            except:
                return 0
        else:
            return 0

    def __getattr__(self, attr):
        if attr == "list":
            return self.as_list()
        elif attr == "list2":
            return self.as_list2()
        elif attr == "dict":
            return self.as_dict()
        elif attr == "string":
            return self.as_string()
        elif attr == "int":
            return self.as_int()
        elif attr == "float":
            return self.as_float()

    def get_string(self, name:str = None) -> str:
        if name:
            if (name in self.dict):
                return Param(self.dict[name]).get_string()
        else:
            # if isinstance(self.param, str):
            #     return self.string
            # elif isinstance(self.param, list):
            #     if len(self.param) > 0:
            #         if isinstance(self.param[0], (str)):
            #             return Param(self.param[0]).string
            if self.param is None:
                return None
            elif isinstance(self.param, list):
                if len(self.param) > 0:
                    if isinstance(self.param[0], (str)):
                        return Param(self.param[0]).string
            elif isinstance(self.param, dict):
                return None
            else:        
                return self.string
        return None

    def get_list(self, name:str = None) -> str:
        if name:
            if (name in self.dict):
                return Param(self.dict[name]).get_list()
        else:
            if self.param is None:
                return None
            elif isinstance(self.param, list):
                return self.param
        return None

    def get_float(self, name:str = None) -> float:
        if name:
            if name in self.dict:
                return Param(self.dict[name]).get_float()
        else:
            if self.param is None:
                return None
            elif isinstance(self.param, dict):
                return None
            elif isinstance(self.param, int) or isinstance(self.param, float):
                return self.float
            elif isinstance(self.param, list):
                if len(self.param) > 0:
                    if isinstance(self.param[0], (int, float)):
                        return Param(self.param[0]).float
        return None

def get_param(value:dict, name:str) -> Param:
    if name in value:
        return Param(value[name])
    else:
        return Param([])




import aiofiles
import yaml
import json
import os
import errno
import dataclasses
import datetime
import decimal
from datetime import date, datetime, timedelta
from typing import Any
import requests
import httpx

import botocore
from boto3.session import Session

from ruamel.yaml.representer import RoundTripRepresenter
from ruamel.yaml import YAML

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

# Перегрузить сохранение YAML файлов на формирование многострочных значений, с использованием "|"
def repr_str(dumper: RoundTripRepresenter, data: str):
    if '\n' in data:
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

# Создать свой объект YAML для сохранения файлов в заданном формате
yamlx = YAML(typ = ['rt', 'string'])
yamlx.representer.add_representer(str, repr_str)

def get_json_data(text:str) -> dict:
    # Заменить переменные окружения в конфиг файле
    for key, value in os.environ.items():
        text = text.replace("${" + f"{key}" + "}", value)
    # Убрать пробелы вокруг текста    
    text = text.strip()    
    # Если пустой файл, то вернуть пустой 
    if text is None:    
        return {}
    elif text[0:1] == "{" or text[0:1] == "}":
        # Распарсит как JSON
        return json.loads(text)
    else:    
        # Распарсить yaml текст настроек
        return yamlx.load(text)
        # return yaml.safe_load(text)

def freads(file_name:str, encoding:str = None):
    """Обычное чтение файла
    """
    with open(file_name, mode = 'r', encoding = encoding if encoding else "utf-8") as f:
        return f.read()

async def aio_reads(file_name:str, encoding:str = None):
    """Асинхронное чтение файла
    """
    async with aiofiles.open(file_name, mode = 'r', encoding = encoding if encoding else "utf-8") as f:
        return await f.read()

def get_json(file_name:str = None, text:str = None) -> dict:
    """Загрузить файл конфигурации и заменить переменные окружения

    :param str config_file: путь файлу конфигурации
    :return dict: словарь опций
    """
    return get_json_data(text = (freads(file_name = file_name) if file_name else text))        

async def aio_get_json(file_name:str = None, text:str = None) -> dict:
    """Загрузить файл конфигурации и заменить переменные окружения

    :param str config_file: путь файлу конфигурации
    :return dict: словарь опций
    """
    return get_json_data(text = (await aio_reads(file_name = file_name) if file_name else text))        

def setenv(name: str, value: str):
    """Записать параметр окружения

    :param str name: наименование параметра
    :param str value: значение параметра
    """
    os.environ[name] = value

def getenv(name: str, default: str = None) -> str:
    """Получить параметр окружения

    :param str name: имя параметра
    :param str default: значение параметра по умолчанию
    :return str: Вовзращаемое значение
    """
    if name:
        value = os.environ.get(name.upper(), None)
        if value:
            return value.replace("\r", "").replace("\n", "").rstrip()
        else:
            return default
    else:
        return None    

def mkdir_path(path):
    """Создать директорий, если не существует
    http://stackoverflow.com/a/600612/190597 (tzot)
    """
    try:
        os.makedirs(path, exist_ok = True)
    except TypeError:
        try:
            os.makedirs(path)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else: raise

def hide_passwords(value: dict, key: str = None) -> dict:
    """Скрыть пароли и токены в словаре

    :param dict value: исходный словарь
    :param str key: Ключ
    :return dict: преобразованный словарь
    """
    if value is None:
        return None
    elif isinstance(value, dict):
        return {k: hide_passwords(v, k) for k, v in value.items()}
    elif isinstance(value, list):
        return [hide_passwords(elem) for elem in value]
    else:
        if key:
            for key_denied in ["PASS", "TOKEN", "ACCESS_KEY", "PWD"]:
                if key_denied in key.upper():
                    return ("<hidden>")
        return (value)        

def my_json_dump(obj: Any, indent: int = None) -> str:
    """Преобразовать в json, сериализировать подчиненные объекты, скрыть пароли

    :param Any obj: исходное значение
    :return str: json как строка
    """
    def default(object_):
        """Функция обработки неизвестных типов
        """
        if isinstance(object_, decimal.Decimal):
            return str(object_)
        elif isinstance(object_, date) or isinstance(object_, datetime):
            return object_.isoformat()
        elif dataclasses.is_dataclass(object_):
            return dataclasses.asdict(object_)
        raise TypeError(f"Don't know how to serialize to json type {type(object_)}")
    return json.dumps(hide_passwords(obj), ensure_ascii=False, default=default, sort_keys=True, indent=indent)

def print_dict_json_indent(value: dict) -> str:
    """Преобразовать словарь в строку json, пароли и токены погасить, убрать пустые значения
    """
    return my_json_dump(obj = value, indent = 2)

def print_dict_json(value: dict) -> str:
    """Преобразовать словарь в строку json, пароли и токены погасить, убрать пустые значения
    """
    return my_json_dump(value)

def print_dict_yaml(value: dict) -> str:
    """Преобразовать словарь в yaml, пароли и токены погасить, убрать пустые значения
    """
    return yaml.dump(hide_passwords(value), allow_unicode=True)

def start_process_timestamp() -> str:
    """Заполнить переменную окружения, старта процесса

    :return str: timestamp старта процесса
    """
    result = getenv("__START_PROCESS_TIMESTAMP__")
    if result is None:
        result = datetime.now().strftime("%Y%m%d %H%M%S")
        setenv("__START_PROCESS_TIMESTAMP__", result)
    return result

def file_temp_path():
    # Получить директорий сохранения файла
    LOG_PATH = getenv("LOG_PATH")
    # Сформировать timestamp файла
    START_PROCESS_TIMESTAMP = start_process_timestamp()
    if LOG_PATH:
        # Получить путь до файла
        file_path = f'{LOG_PATH}/{START_PROCESS_TIMESTAMP}'
        # Создать директорий, если еще нет
        mkdir_path(file_path)
        return file_path
    else:
        return None

def save_dict(value_dict: dict, filename: str, desc: str = None, tp: str = "yaml"):
    """Сохранить словарь как YAML файл в локальной папке или в S3. Зависит от параметров окружения. Если значение не выгружено ни туда ни туда, то отправим в info лог как json !!!

    :param dict value_dict: сохраняемый словарь
    :param str filename: имя файла
    :param str desc: описание словаря
    :param str tp: Тип сохранения как "json" или "yaml"
    """
    if tp.lower() == "yaml":
        _filename = filename if ".yaml" in filename.lower() else f'{filename}.yaml'
        # Функция преобразования в строку
        dict2str = print_dict_yaml
    elif tp.lower() == "json":
        _filename = filename if ".json" in filename.lower() else f'{filename}.json'
        # Функция преобразования в строку
        dict2str = print_dict_json_indent
    else:
        raise Exception(f'Неизвестный тип преобразования "{tp}" словаря')    
    # Получить директорий сохранения файла
    LOG_PATH = getenv("LOG_PATH")
    # Сформировать timestamp файла
    START_PROCESS_TIMESTAMP = start_process_timestamp()
    is_save = False
    if LOG_PATH:
        # Получить путь до файла
        file_path = f'{LOG_PATH}/{START_PROCESS_TIMESTAMP}'
        # Создать директорий, если еще нет
        mkdir_path(file_path)
        # Если задан директорий, то в нем создать yaml файл
        full_path = f'{file_path}/{_filename}'
        open(full_path, "w", encoding = "utf-8").write(dict2str(value_dict))
        logger.info(f'save as {tp.lower()} into file: "{full_path}')
        is_save = True
    if getenv("S3_URL", None):    
        # Если, заданы параметры хранилища S3
        # Установить сессию с S3
        session = Session(
            aws_access_key_id = getenv("S3_ACCESS_KEY_ID", None), 
            aws_secret_access_key = getenv("S3_SECRET_ACCESS_KEY", None))
        s3 = session.resource('s3', endpoint_url = getenv("S3_URL", None))
        # Получить имя корзины
        backet_name = getenv("S3_BUCKET_NAME", None)
        # Сохранить данные
        s3.Bucket(backet_name).put_object(
            Key = f'{START_PROCESS_TIMESTAMP}/{_filename}', 
            Body = dict2str(value_dict),
            ContentDisposition = desc or '',
            ContentType = "text/x-yaml")
        logger.info(f'save as {tp.lower()} into s3: "{backet_name}/{START_PROCESS_TIMESTAMP}/{_filename}"')
        is_save = True
    return is_save    

def save_object(value_dict: dict, filename: str, desc: str = None, log_func = None, tp = None):
    """Сохранить словарь как YAML файл в локальной папке или в S3. Зависит от параметров окружения. Если значение не выгружено ни туда ни туда, то отправим в info лог как json !!!

    :param dict value_dict: сохраняемый словарь
    :param str filename: имя файла
    :param str desc: описание словаря
    :param log_func: функция логирования, если словарь никуда не сохранен
    """
    if not save_dict(value_dict = value_dict, filename = filename, desc = desc, tp = tp or getenv("LOG_TABLE_FORMAT", "yaml")):    
        # Если таблица никуда не выгружена, то отправим в лог !!!
        if log_func:
            # Добавим описание в скобочки
            _desc = f'({desc})' if desc else ""
            log_func(f' {filename}{_desc}: {print_dict_json(value_dict)}')

async def telegram(message: str, token: str = None, chats: list = None, success: bool = None):
    """Отправить сообщение в телеграм

    :param str parse_message: отправляемое сообщение
    :param str token: токен для бота, defaults to None
    :param str chats: массив чатов, defaults to None
    :param bool success: отправить хороший или плохой смайлик
    """
    _token = token or getenv("TELEGRAM_TOKEN")
    if _token:
        for chat_id in (chats or [getenv("TELEGRAM_CHAT")]):
            if chat_id:
                emoji = ('%F0%9F%98%83' if success else '%F0%9F%98%A1') if success is not None else None
                for message in [emoji, message]:
                    if message is not None:
                        async with httpx.AsyncClient() as client:
                            try:
                                telegram_url = f"https://api.telegram.org/bot{_token}/sendMessage"
                                telegram_param = {
                                    "chat_id": chat_id,
                                    "text": message,
                                    "parse_mode": "HTML",
                                }    
                                response =  await client.get(f"{telegram_url}", params = telegram_param)
                                # Если код возврата, ошибка, то поднять ошибку
                                response.raise_for_status()
                                logger.debug(f'send message to telegram')
                            except httpx.HTTPError as exc:
                                # Не поднимать ошибку, если нет доступа к телеграмму
                                logger.error(f'HTTP Exception for "{exc.request.url}" - "{exc}"')


async def teams(message: str, teams_url: str = None, success: bool = None):
    """Отправить сообщение в teams

    :param str parse_message: сообщение
    :param str teams_url: путь до teams, defaults to None
    :param bool success: показать смайлик, defaults to None
    """
    _teams_url = teams_url or getenv("TEAMS_WEBHOOK")
    if _teams_url:
        logger.debug(f'send message in teams path: "{_teams_url}"')
        body = []
        if success is not None:
            body.append({
                "horizontalAlignment": "left",
                "type": "TextBlock",
                "text": "🙂" if success else "🙁",
                "size": "extraLarge",
                "weight": "bolder"})

        body.append({
            "type": "FactSet",
            "facts": [ {"value": line} for line in message.split("\n")]
        })
        # Сформировать сообщение
        js_message = {
            "type": "message",
            "attachments": [
                {
                    "contentType": "application/vnd.microsoft.card.adaptive",
                    "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
                    "version": "1.0",
                    "content": {
                        "type": "AdaptiveCard",
                        "padding": "none",
                        "body": body
                    }
                }
            ]
        }     
        # Отправить сообщение               
        async with httpx.AsyncClient() as client:
            try:
                response =  await client.post(f"{_teams_url}", json = js_message)
                # Если код возврата, ошибка, то поднять ошибку
                response.raise_for_status()
                logger.debug(f'send message to teams')
            except httpx.HTTPError as exc:
                # Не поднимать ошибку, если нет доступа к teams
                logger.error(f'HTTP Exception for "{exc.request.url}" - "{exc}"')

async def send_message(message: str = None, success: int = None, service: str = None):
    """Отправить сообщение пользователю, в телеграм или teams

    :param str message: сообщение, defaults to None
    :param int success: показать упешный или не успешный смайли, defaults to None
    """
    _message = message
    if _message:
        if service:
            _message = "\n".join([
                f'{service}', 
                "===================================", 
                _message])
        # Отправить сообщение в telegram       
        await telegram(message = _message, success = success)
        # Отправить сообщение в teams
        await teams(message = _message, success = success)
        if success is None or success:
            logger.info(_message.replace("\n", " "))
        else:    
            logger.error(_message.replace("\n", " "))



import asyncio

from paqlang.utils import aio_reads
from .gitlab_opers import git_pool

import cx_Oracle_async
import cx_Oracle

import logging
logger = logging.getLogger(__name__)

def split_blocks(text:str):
    """Разбить запрос на блоки выполнения по символу /
    """
    result = []
    delim = "-----/----EXECUTE-----/----"
    text2 = "\n".join([delim if line.strip() == "/" else line for line in text.split("\n")])
    # text4 = "\n".join([delim if line.strip() == ";" else line for line in text2.split("\n")])
    for code in text2.split(delim):
        if code.strip():
            result.append(code)
    return result        

class Pool:

    pools = None

    def __init__(self):
        self.pools = {}

    async def get(self, user, password, dsn):
        # Получить ключ pool
        key = f'{user=}::{password=}::{dsn=}'
        if key not in self.pools:
            self.pools[key] = await cx_Oracle_async.create_pool(
                    user = user, 
                    password = password, 
                    dsn = dsn,
                    min = 2,
                    max = 10)
        return self.pools[key]

pool = Pool()

class OracleOpers:

    async def execute(pgm, param, p_queue, in_queue = None, out_queue = None, fetch = None):
        if len(in_queue):
            # Получить подключение из pool коннекций
            oracle_pool = await pool.get(
                user = param.get_string("oracle_user"),
                password = param.get_string("oracle_password"),
                dsn = param.get_string("oracle_dsn"))
            # Если указано в параметрах подключение  gitlab, то подключиться к pool git
            gl = await git_pool.get_project(param = param) if "git_url" in param.dict else None
            # Провалится до курсора
            async with oracle_pool.acquire() as connection:
                async with connection.cursor() as cursor:
                    # Пока входная очередь запросов или файлов не пустая
                    while len(in_queue):
                        # Получить первое значение из очереди
                        in_param = in_queue.pop(0)
                        if gl:
                            # Работаем с gitlab
                            try:
                                # Прочитать файл из git, передаем имя файла
                                file = await gl.get_file(param, in_param)
                            except Exception as e:
                                logger.error(f'get file: {in_param} - {e}')
                                raise    
                            # Имя файла
                            fname = file["path"]
                            # Содержимое файла
                            sql = file["text"]
                        else:   
                            # Если задана атрибут path, значит считать файл с локального диска 
                            if "path" in param.dict:
                                # Возможность задать кодировку файла
                                encoding = param.get_string("encoding") or "utf-8"
                                # 
                                path = param.get_string("path")
                                # Если path содержит, то сделать подстановку, иначе подставить имя файла как есть в очереди
                                fname = path.replace(":1", in_param) if path and (":1" in path) else in_param
                                # Прочитать содержимое файла
                                sql = await aio_reads(file_name = fname, encoding = encoding)
                            else:
                                # Запрос из потока, не из файла данных
                                fname = None
                                sql = in_param    
                        if fetch == "all":
                            # Получить все записи из запроса
                            await cursor.execute(sql)
                            columns = [col[0].lower() for col in cursor._cursor.description]
                            cursor._cursor.rowfactory = lambda *args: dict(zip(columns, args))
                            for row in await cursor.fetchall():
                                out_queue.append(row)
                        elif fetch == "one":
                            # Получить первую запись из запроса
                            await cursor.execute(sql)
                            columns = [col[0].lower() for col in cursor._cursor.description]
                            cursor._cursor.rowfactory = lambda *args: dict(zip(columns, args))
                            row = await cursor.fetchone()
                            if row:
                                out_queue.append(row)
                        else:
                            # Включить dbms_output
                            await cursor.callproc("dbms_output.enable")
                            try:
                                # Разбить запрос на подзапросы и выполнить каждый
                                for code in split_blocks(sql):
                                    logger.info(f'{code=}')
                                    await cursor.execute(code)
                                # Подготовить dbms_output
                                dbms_output = []
                                # Размер чтения порции строк из dbms_output
                                chunk_size = 100
                                # Создать преременные для чтения
                                lines_var = cursor._cursor.arrayvar(str, chunk_size)
                                num_lines_var = cursor._cursor.var(int)
                                num_lines_var.setvalue(0, chunk_size)
                                # Прочитать порциями dbms_output
                                while True:
                                    await cursor.callproc("dbms_output.get_lines", (lines_var, num_lines_var))
                                    num_lines = num_lines_var.getvalue()
                                    lines = lines_var.getvalue()[:num_lines]
                                    for line in lines:
                                        dbms_output.append(line or "")
                                    if num_lines < chunk_size:
                                        break
                                # Вернуть все хорошо, с dbms_output    
                                result = {
                                    "result": 1,
                                    "output": '\n'.join(dbms_output)
                                }    
                            except  cx_Oracle.DatabaseError as e:            
                                # Заполнить ошибку выполнения
                                error, = e.args
                                result = {
                                    "errmsg": error.message,
                                    "result": 0
                                }
                            # Если запрос из файла, то добавить в результат имени файла
                            if fname:
                                logger.info(f'oracle exec: "{fname}"')
                                result["file"] = fname     
                            out_queue.append(result)

    async def multiple_oracle_execute(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Выполнить SQL запрос. Запросы разбиваются на подзапросы "/" и ";"

        oracle_user:str - Подключение к oracle
        oracle_password:str
        oracle_dsn:str

        Входная очередь:
        - если не заданы параметры, то список текстов запросов. 
        - если git_url:str, то список файлов из getlab. Файл загружается из git и выполняется. 
          Необходимы также git_token, git_repo, git_branch
        - если path:str, то список файлов из директория на диске. 
          Формат атрибута "...:1...". Вместо :1 подставляется заданный файл.
        
        Результат массив выполенных запросов
        {"result": 0 или 1, # 0 - error execute, 1 - succes execute
         "output":str, # строки dbms_output потока
         "errmsg":str, # ошибка выполнения, если result = 0
         "file":str, # путь к файлу, если запрос был загружен из файла или git
        }
        """
        await OracleOpers.execute(pgm, param, p_queue, in_queue, out_queue, fetch = False)
        return ['success']

    async def multiple_oracle_fetchall(pgm, param, p_queue, in_queue = None, out_queue = None):
        await OracleOpers.execute(pgm, param, p_queue, in_queue, out_queue, fetch = "all")
        return ['success']

    async def multiple_oracle_fetchone(pgm, param, p_queue, in_queue = None, out_queue = None):
        await OracleOpers.execute(pgm, param, p_queue, in_queue, out_queue, fetch = "one")
        return ['success']



import datetime

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class DateOpers:

    async def single_now(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Текущая дата в ISO формате, как строка

        """
        out_queue.append(datetime.datetime.now().isoformat())
        return ['success']



import asyncio
import re
import httpx
import urllib
import chardet
import codecs
from datetime import date, datetime

import logging
logger = logging.getLogger(__name__)

class GitlabAsync:

    project = None
    
    async def httpx_get(self, url, raw = False):
        # logger.info(f'get: {url}')
        async with httpx.AsyncClient() as client:
            try:
                response =  await client.get(f'{url}', headers = {'Private-Token': self.token})
                # Если код возврата, ошибка, то поднять ошибку
                response.raise_for_status()
                if raw:
                    return response.content
                else:
                    return response.json()
            except httpx.HTTPError as exc:
                # Не поднимать ошибку, если нет доступа к teams
                logger.error(f'HTTP Exception for "{exc.request.url}" - "{exc}"')
                raise

    async def build_url(self, resources, **kwargs):
        url = []
        # Список проектов
        url += ["/projects" if resources.get("projects") else ""]
        # Проект зафиксирован
        url += [f"/projects/{resources.get('project_id')}" if resources.get("project_id") else ""]
        # Список веток
        url += ["/repository/branches" if resources.get("branches") else ""]
        url += ["/repository/commits" if resources.get("commits") else ""]
        # Список файлов
        url += ["/repository/tree" if resources.get("tree") else ""]
        # Спсиок tags
        url += ["/repository/tags" if resources.get("tags") else ""]
        # Получить содержимое файла
        url += [f"/repository/files/{resources.get('raw')}/raw" if resources.get("raw") else ""]
        # Получить diff коммита
        url += [f"/repository/commits/{resources.get('diff')}/diff" if resources.get("diff") else ""]
        # Получить параметры api, None не включать
        kwargs2 = urllib.parse.urlencode({key: value for key, value in kwargs.items() if value is not None})
        parameters = f"{kwargs2}"
        # Сформировать url строку запроса
        url_string = f"{self.url}/api/v4{''.join(url)}"
        # Вернуть строку запроса с параметрами
        return f"{url_string}?{parameters}" if parameters else url_string

    async def set_project(self, url, token, name):
        """Найти проект в репозитории и запомнить в классе

        :param str url: путь к gitlab
        :param str token: токен
        :param str name: имя репозитория
        """
        self.url = url
        self.token = token
        for project in await self.httpx_get(await self.build_url(resources = {"projects": True}, search = name)):
            self.project = project

    async def get_raw(self, resources, **kwarg):
        return await self.httpx_get(await self.build_url(resources = resources, **kwarg), raw = True)

    async def get_pages(self, resources, **kwarg):
        page = 0
        per_page = 50
        result = []
        while 1:
            page += 1
            records = await self.httpx_get(await self.build_url(resources = resources, **kwarg, page = page, per_page = per_page, pagination = "legacy"))
            if records:
                result.extend(records)
            if len(records) < per_page:
                break    
        return result

    async def get_branches(self, param):
        return await self.get_pages(
            resources = {
                "project_id": self.project["id"], 
                "branches": True}, 
            regex = param.get_string("regex"),
            search = param.get_string("search"))

    async def get_projects(self, param):
        result = []
        regex = param.get_string("regex")
        for project in await self.get_pages(
            resources = {
                "projects": True}, 
            search = param.get_string("search")):
            if regex and not re.match(regex, project["name"]):
                continue
            result.append(project["name"])
        return result    


    async def get_tags(self, param):
        result = []
        regex = param.get_string("regex")
        for tag in await self.get_pages(
            resources = {
                "project_id": self.project["id"], 
                "tags": True}, 
            search = param.get_string("search")):
            if regex and not re.match(regex, tag["name"]):
                continue
            result.append(tag["name"])
        return result    

# https://gitlabci.dpd.ru/api/v4/projects/260/repository/tree?id=260&iterator=true&page=2&pagination=legacy&path=Views%2F&per_page=20&recursive=true&ref=tests%2F2.133.0.0
    async def get_tree(self, param, path):
        result = []
        regex = param.get_string("regex")
        for file in await self.get_pages(
            resources = {
                "project_id": self.project["id"], 
                "tree": True}, 
            path = path, 
            recursive = "true",
            ref = param.get_string("git_branch")):
            if file["type"] == "tree":
                continue
            if regex and not re.match(regex, file["path"]):
                continue
            result.append({"path": file["path"], "name": file["name"]})
        return result

#https://gitlabci.dpd.ru/api/v4/projects/260/repository/files/!release%2FDSORT2-439%2F_source_upd%2Esql/raw?ref=tests%2F2.133.0.0 "HTTP/1.1 200 OK"
#https://gitlabci.dpd.ru/api/v4/projects/260/repository/files/Materialized+view+logs%2Farg_state%2Esql/raw?ref=tests%2F2.133.0.0'
    async def get_file(self, param, path):
        branch = param.get_string("git_branch")
        file_bytes = await self.get_raw(
            resources = {"project_id": self.project["id"], 
                         "raw": path.replace("/", "%2F").replace(".", "%2E")}, 
            ref = branch)
        # Определить кодировку файла
        file_encoding = chardet.detect(file_bytes)["encoding"]
        file_encoding = file_encoding.upper() if file_encoding else None
        # Определение формата входного файла и BOM
        # Переменная для определения BOM для UTF8 фалов
        bom = 0
        # Формат входного файла
        encode = "cp1251"
        if file_encoding in ["UTF-8-SIG", "UTF-8", "MACROMAN"]:
            encode = "utf-8"
            if file_bytes[0] == 239:  # Первый байт BOM
                bom = 1
        # Раскодировать входную байтовую строку, во внутренний формат
        file_str = ""                        
        encoding = 0
        try:
            file_str = codecs.decode(file_bytes, encode)
            encoding = 1
        except Exception:
            pass
        return {
            "branch": branch,
            "path": path,
            "encode_detect": file_encoding,
            "encode": encode,
            "bom": bom,
            "encoding": encoding,
            "text": file_str
        }    

    async def get_commits(self, param, branch):
        result = []
        for commit in await self.get_pages(
            resources = {
                "project_id": self.project["id"], 
                "commits": True}, 
            ref_name = branch, 
            since = date.fromisoformat(param.get_string("since")).isoformat() if param.get_string("since") else None):
            result.append(commit)
        return result

    async def get_diff(self, param, sha):
        result = []
        for diff in await self.get_pages(
            resources = {
                "project_id": self.project["id"], 
                "diff": sha}):
            result.append(diff["new_path"])
        return result


class Pool:

    pools = None

    def __init__(self):
        self.pools = {}

    async def get_project(self, param):
        url = param.get_string("git_url")
        token = param.get_string("git_token")
        repo = param.get_string("git_repo")
        key = f'{url=}::{token=}::{repo=}'
        if key not in self.pools:
            # Подключиться к репозиторию
            gl = GitlabAsync()
            await gl.set_project(url = url, token = token, name = repo)
            # Подключиться к проекту
            self.pools[key] = gl
        # logger.info(f'gitlab pool : {repo}@{url}')    
        return self.pools[key]
        
git_pool = Pool()  



from datetime import datetime

from .aio_gitlab import git_pool

import logging
logger = logging.getLogger(__name__)

def str2date(date:str) -> datetime:
    """Преобразовать строку с датой в поле даты, отбросить временную зону

    :param str date: строка даты
    :return datetime: результат как дата
    """
    return datetime.fromisoformat(date).replace(tzinfo=None)

def my_strip(s: str) -> str:
    """Убрать переводы строк и сжать строку справа
    """
    return ' '.join(s.split()).rstrip()

class GitlabOpers:

    async def single_gitlab_projects(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список проектов в репозитории

        git_url:str - Подключение к gitlab
        git_token:str
        search:str = None - наименование проекта
        regex:str = None - regex шаблон отбора веток

        Входная очередь - не используется
        
        Результат массив проектов
        """
        gl = await git_pool.get_project(param = param)
        out_queue.extend(await gl.get_projects(param = param))
        return ['success']

    async def single_gitlab_branches(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список веток в репозитории

        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        search:str = None - наименование ветки
        regex:str = None - regex шаблон отбора веток

        Входная очередь - не используется
        
        Результат массив веток
        """
        gl = await git_pool.get_project(param = param)
        for branch in await gl.get_branches(param):
            out_queue.append(branch["name"])
        return ['success']

    async def multiple_gitlab_walk(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список файлов в репозитории

        tasks:int = None - Максимальное кол-во потоков обработки операций
        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        git_branch:str - ветка
        regex:str = None - regex шаблон отбора файлов

        Входная очередь, содержит список корневых узлов, от которых запускается процесс
        Все файлы в репозитории идут от "/"
        
        Результат массив
        - name: - имя файла
          path: - полный маршрут файла
        """
        if len(in_queue):
            gl = await git_pool.get_project(param = param)
        while len(in_queue):
            # Получить список файлов
            out_queue.extend(await gl.get_tree(param = param, path = in_queue.pop(0)))
        return ['success']

    async def multiple_gitlab_freads(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить содержимое файлов в репозитории

        tasks:int = None - Максимальное кол-во потоков обработки операций
        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        git_branch:str - ветка

        Входная очередь, содержит список файлов, полный маршрут от корня

        Результат массив
        - bom: 0
          branch: - ветка
          encode: cp1251 - найденная кодировка
          encode_detect: ASCII 
          encoding: 1 - файл раскодирован 
          path: - полный путь файла
          text: - содержимое файла 
        """
        if len(in_queue):
            gl = await git_pool.get_project(param = param)
            while len(in_queue):
                # Получить файл
                out_queue.append(await gl.get_file(param, in_queue.pop(0)))
        return ['success']


    async def single_gitlab_tags(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список тегов в репозитории

        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        search:str = None - наименование тега
        regex:str = None - regex шаблон отбора тегов

        Входная очередь - не используется
        
        Результат массив тегов
        """
        gl = await git_pool.get_project(param = param)
        out_queue.extend(await gl.get_tags(param))
        return ['success']


    async def multiple_gitlab_commits(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список тегов в репозитории

        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        
        Результат массив тегов
        """
        if len(in_queue):
            gl = await git_pool.get_project(param = param)
            while len(in_queue):
                # Получить ветку
                branch = in_queue.pop(0)
                if isinstance(branch, str) and len(branch) > 0:
                    for commit in await gl.get_commits(param, branch):
                        out_queue.append({
                            "branch": branch,
                            "id": commit["id"],
                            "last_committed_date": commit["committed_date"], # Commit, при rebase не изменяется
                            "last_committed_date2": str2date(commit["committed_date"]),
                            "committer_name": commit["committer_name"],
                            "authored_date": commit["authored_date"], # Commit, при rebase не изменяется
                            "authored_date2": str2date(commit["authored_date"]), # Commit, при rebase не изменяется
                            "author_name": commit["author_name"],
                            "last_message": my_strip(commit["message"]),
                            # "diff": await gl.get_diff(param, commit["id"]),
                        })
        return ['success']

    async def multiple_gitlab_diff(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список измененых файлов по SHA commit

        tasks:int = None - кол-во потоков
        git_url:str - Подключение к gitlab
        git_token:str
        git_repo:str - проект репозитория
        
        Входная очередь массив sha или {"id":}
        Результат массив новых файлов
        """
        if len(in_queue):
            gl = await git_pool.get_project(param = param)
            while len(in_queue):
                # Получить ветку
                item = in_queue.pop(0)
                sha = item if isinstance(item, str) else item.get("id")
                if sha:
                    out_queue.extend(await gl.get_diff(param, sha))
        return ['success']



import asyncio
import re
import datetime

from .util_opers import coalesce, get_condition
from ..param import Param, get_param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)


class ControlOpers:

    async def single_exit(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Завершить выполнение программы
        """
        return ['exit']


    async def single_when(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Если условие истино, ты выполнить stage

        Входное значение, если указано "in", иначе входная очередь
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        Если условие не задано, то выполняется, если входная очередб не пустая
        """
        # Скопировать очередь
        out_queue.extend(in_queue)
        if len(param.list) == 0:
            result = len(in_queue) != 0
        else:    
            test_value = coalesce([param.get_float("in"), param.get_string("in"), p_queue.get_float(), p_queue.get_string()])
            result = get_condition(test_value, param)
        if result:
            return ['success']
        else:
            # Откатить стайдж
            return ['when']

    async def single_break(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Если условие истино, ты прервать цикл while (break)

        Входное значение, если указано "in", иначе входная очередь
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        """
        # Скопировать очередь
        out_queue.extend(in_queue)
        if len(param.list) == 0:
            result = len(in_queue) == 0
        else:    
            test_value = coalesce([param.get_float("in"), param.get_string("in"), p_queue.get_float(), p_queue.get_string()])
            result = get_condition(test_value, param)
        if result:
            return ['break']
        else:
            return ['success']

    async def single_continue(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Если условие истино, ты прервать цикл while (continue)

        Входное значение, если указано "in", иначе входная очередь
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        """
        # Скопировать очередь
        out_queue.extend(in_queue)
        test_value = coalesce([param.get_float("in"), param.get_string("in"), p_queue.get_float(), p_queue.get_string()])
        if get_condition(test_value, param):
            return ['continue']
        else:
            return ['success']



import logging
import re
# Установить текущи логгер
logger = logging.getLogger(__name__)

from .util_opers import get_attr

class StringOpers:

    async def single_search(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Отобрать элементы из очереди, удовлетворяющие условию

        in_queue - входная очередь
        attr:str - если указано, то анализировать атрибут из элемента очереди
        regex:str или param:str - условия отбора
        """
        attr = param.get_string("attr")
        regex = param.get_string("regex") or param.get_string()
        assert regex is not None, 'Не задан атрибут поиска "regex"'
        for elem in in_queue:
            m = re.match(regex, get_attr(attr, elem))
            if m:
                out_queue.append(elem)
        return ['success']

    async def single_match(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вырезать из элемента, заданный regex подстроку

        in_queue - входная очередь
        regex:str или param:str - условие получения подстроки, группа 1
        """
        regex = param.get_string("regex") or param.get_string()
        assert regex is not None, 'Не задан атрибут поиска "regex"'
        for elem in in_queue:
            m = re.match(regex, elem)
            if m:
                out_queue.append(m.group(1))
        return ['success']

    async def single_replace(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Заменить во входной очереди, в строках, заданный regex на подстроку

        in_queue - входная очередь
        regex:str или param:str - условие получения подстроки
        dest:str - на что заменять, если не указано, заменяем на пустую строку
        """
        regex = param.get_string("regex") or param.get_string()
        dest = param.get_string("dest") or ""
        assert regex is not None, 'Не задан атрибут поиска "regex"'
        for text in in_queue:
            out_queue.append(re.sub(regex, dest, text))
        return ['success']

    async def single_subst(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Заменить в заданной строке :1, на элементы из входной очереди

        in_queue - входная очередь
        text:str или param:str - строка шаблон
        """
        text = param.get_string("text") or param.get_string()
        assert text is not None, 'Не задана исходная строка - шаблон'
        for elem in in_queue:
            out_queue.append(text.replace(":1", str(elem)))
        return ['success']

    async def single_split(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Разбить строку на элементы

        param:str - строка разбиения, по умолчанию \n"""
        attr = param.get_string("attr")
        split_str = param.get_string() or '\n'
        while len(in_queue):
            # Получить файл
            text = in_queue.pop(0)
            # Распарсить атрибут 
            text = get_attr(attr = attr, elem = text, is_raise = False) if attr else text
            if isinstance(text, str):
                #logger.info(f'{text=}')
                pass
            else:
                logger.error(f'{text=}')    
            # Проскочить заданные объекты    
            for line in text.split(split_str):
                _line = line.replace("\\", "/").strip()
                if len(_line):
                    out_queue.append(_line)
        return ['success']

    async def single_join(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Соединить элементы массива в строку

        param:str - строка объединения, по умолчанию \n
        """
        join_str = param.get_string() or '\n'
        out_queue.append(join_str.join([f'{elem}' for elem in in_queue if len(f'{elem}') > 0]))
        return ['success']

    async def single_transform(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Найти в списке элемент, независимо от регистра и вернуть правильное наименование

        """
        transform = {}
        for elem in param.list:
            if isinstance(elem, str):
                transform[elem.lower()] = elem
        for elem in in_queue:        
            if isinstance(elem, str):
                if elem.lower() in transform:
                    out_queue.append(transform[elem.lower()])
        return ['success']



from .utils import save_object
import asyncio
import re
import datetime

from .param import Param
from .data import Data, mems
from .functions.other_opers import OtherOpers
from .functions.math_opers import MathOpers
from .functions.string_opers import StringOpers
from .functions.date_opers import DateOpers
from .functions.queue_opers import QueueOpers
from .functions.bool_opers import BoolOpers
from .functions.control_opers import ControlOpers
from .functions.io_opers import IoOpers
from .functions.dict_opers import DictOpers

import inspect

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

def log(mess:str):
    # stage_desc = f'{datetime.datetime.now()}:{mess}'
    stage_desc = f'{mess}'
    logger.debug(f'{stage_desc}')


class Pgm:
    """Класс выполнения программы
    """
    # Внешние процедуры параллельной обработки
    multiple_opers = None
    # Список внутренних процедур обработки
    single_opers = None
    # Список подпрограмм
    macros = None
    #
    result = None

    def __init__(self):
        """Создать объект выполняемая программа

        :param str text: текст программы на языке YAML, defaults to None
        """
        # Получить список внутренних процедур программы
        self.multiple_opers = {}
        # Список внутренних процедур обработки
        self.single_opers = {}
        # Список подпрограмм
        self.macros = {}
        for cls in [
            OtherOpers, 
            MathOpers, 
            StringOpers, 
            DateOpers, 
            QueueOpers, 
            ControlOpers, 
            IoOpers,
            BoolOpers,
            DictOpers]:
            self.append_opers(in_class = cls)

    def append_macros(self, macros:dict):
        if macros:
            for key, value in macros.items():
                self.macros[key] = value

    def append_opers(self, in_class:object):
        if in_class:
            for fullname, coroutine in inspect.getmembers(in_class, predicate = inspect.isfunction):
                if fullname[0:1] == "_":
                    continue
                name = fullname.replace("multiple_", "").replace("single_", "")
                value = {
                    "group": in_class.__name__,
                    "coroutine": coroutine,
                    "doc": coroutine.__doc__,
                    "name": name,
                    "fullname": fullname,
                    "multiple": "multiple_" in fullname,
                }
                if value["multiple"]:
                    self.multiple_opers[name] = value
                else:       
                    self.single_opers[name] = value
    
    def get_data(self, name: str):
        """Вернуть результат выполнения заданного stage

        :param str stage_name: наименование stage
        :return list: данные stage
        """
        return mems.get_data(name = name)
        
    def set_data(self, name:str, data:list):
        """Запомнить результат выполнения stage

        :param str stage_name: наименование stage
        :param list data: результат выполнения stage
        """
        mems.set_data(name = name, value = data)

    def get_lines(self, in_list: list):
        out_list = []
        for text in in_list:
            # Проскочить заданные объекты    
            for line in text.split("\n"):
                _line = line.replace("\\", "/").strip()
                if len(_line):
                    out_list.append(_line)
        return out_list            

    def context(self, result, queue, date_start, stack = None):
        return {"result": result, "queue": queue, "start": date_start, "stop": datetime.datetime.now(), "stack": stack}

    async def function(self, stage_name:str, coroutine:object, pgm: dict, param: Param, p_queue: Param, in_queue: list, backtrace:list):
        out_queue = []
        date_start = datetime.datetime.now()
        _backtrace = " -> ".join(backtrace)
        try:
            # logger.info(f'++ "{_backtrace}" : in: "{len(in_queue)}"')
            result = await coroutine(pgm = pgm, param = param, p_queue = p_queue, in_queue = in_queue, out_queue = out_queue)
            # logger.info(f'-- "{_backtrace}": "{result}" out: "{len(out_queue)}"')
        except Exception as e:
            logger.error(f'backtraces: {" -> ".join(backtrace)}, error: {e}')    
            raise
        return self.context(result = result, queue = out_queue, date_start = date_start)

    async def execute(self, pgm:dict, in_queue:list, step:int = 0, options:dict = None, backtrace:list = None) -> list:
        """Выполнить блок программы

        :param dict pgm: блок программы
        :param list in_queue: входная очередь
        :param list out_queue: выходная очередь
        :param dict sub: массив библиотек
        :return bool: Если истина, то ветка выполняется дальше, иначе прерывается
        """
        # Операция выполнена нормально
        result = []
        stack_list = []
        stack_dict = {}
        out_queue = []
        start_date = datetime.datetime.now()
        if options and options.get("while"):
            # logger.info(f'while: {pgm=}')
            # Выполнить задачи в цикле
            __out_queue = []
            i = 0
            while 1:
                # Создать входную и выходную очередь для задач
                __in_queue = []
                # Скопировать первый раз входную очередь, далее результат предыдущего блока
                __in_queue.extend(in_queue if __out_queue is None else __out_queue)
                name = f'while'
                # Выполнить пакет задач
                _result = await self.execute(pgm = pgm, in_queue = __in_queue, step = step, backtrace = backtrace +[name])
                i += 1
                stack_list.append({f'iter {i}': _result})
                # Прервать выполнение далее блоков программы
                # break
                if 'break' in _result["result"]:
                    break
                # continue
                if 'continue' in _result["result"]:
                    continue
                __out_queue = _result["queue"]
            # Вернуть последнюю очередь, как результат, если все блоки были выполнены
            out_queue.extend(__out_queue)    
            result = ['success']
        elif isinstance(pgm, str):
            # Блок программы строка, сформируем словарь из команды и пустого параметра и рекурсивно вызовем себя
            _result = await self.execute(pgm = {pgm: None}, in_queue = in_queue, step = step, backtrace = backtrace)
            return _result
        elif isinstance(pgm, dict):
            # Блок программы словарь, если несколько задая, то выполнить параллельно
            # Создать входную и выходную очередь
            __in_queue = []
            __in_queue.extend(in_queue)
            __out_queue = []
            # Массив задач
            async_tasks = []
            # Сформируем список параллельно выполняемых блоков
            stages = ",".join([key for key in pgm])
            # Перебрать параллельные блоки 
            for stage_name, stage_data in pgm.items():
                # stage_name - наименование блока или функции
                # stage_data - параметры функции или тело блока
                if stage_name == "call":
                    # Вызвать подпрограмму из библиотеки, рекурсивно выполнить код функции
                    # Функция есть в классе Opers внутренних функций обработчика
                    __out_queue = []
                    name = f'{stage_name}'
                    async_tasks.append({"name": name, "out": __out_queue, "func": asyncio.create_task(self.execute(pgm = self.macros[Param(stage_data).string], in_queue = __in_queue, step = step + 1, backtrace = backtrace + [name]))})
                elif stage_name in self.single_opers:
                    # Функция есть в классе Opers внутренних функций обработчика
                    # Создать асинхронную задачу по функции и добавить в спсиок задач
                    # Param(stage_data) - параметры функции
                    # Param(__in_queue) - получить параметры входной очеререди
                    name = f'{stage_name}'
                    async_tasks.append({"name": name, "param": Param(stage_data).list, "out": __out_queue, "func": asyncio.create_task(self.function(stage_name = name, coroutine = self.single_opers[stage_name]["coroutine"], pgm = self, param = Param(stage_data), p_queue = Param(__in_queue), in_queue = __in_queue, backtrace = backtrace + [name]))})
                elif stage_name in self.multiple_opers:
                    # Функция есть в классе Opers внутренних функций обработчика
                    # Получить функцию 
                    # Создать асинхронную задачу по функции и добавить в спсиок задач
                    # Кол-во задач
                    task_count = 10 if len(__in_queue) > 10 else len(__in_queue)
                    # Param(stage_data) - параметры функции
                    # Param(__in_queue) - получить параметры входной очеререди
                    params = Param(stage_data)
                    if params.dict:
                        if params.get_float("tasks"):
                            task_count = int(params.get_float("tasks"))
                    for i in range(0, task_count):
                        name = f'{stage_name}[{i}]'
                        async_tasks.append({"name": name, "param": params.list, "out": __out_queue, "func": asyncio.create_task(self.function(stage_name = name, coroutine = self.multiple_opers[stage_name]["coroutine"], pgm = self, param = params, p_queue = Param(__in_queue), in_queue = __in_queue, backtrace = backtrace + [name]))})
                else:
                    # Если функция не найдена, то считаем, что пришел блок программы и создаем асинхронную задачу на его выполнение
                    name = f'{stage_name}'
                    async_tasks.append({"stage": 1, "out": __out_queue, "name": name, "func": asyncio.create_task(self.execute(pgm = stage_data, in_queue = __in_queue, step = step + 1, options = {"while": "while" in stage_name}, backtrace = backtrace + [name]))})
            # Выполнить пул задач
            # Время старта блока
            for task in async_tasks:
                _result = await task["func"]
                # if "st1" == task["name"]:
                #     logger.info(f'st1: {_result}')
                # "success" in _result
                if "stage" in task:
                    mems.set_data(task["name"], _result["queue"])
                if 'success' in _result["result"]:
                    out_queue.extend(_result["queue"])    
                if "param" in task:    
                    _result["param"] = task["param"]
                stack_dict[f'{task["name"]}'] = _result
                try:
                    result = result + _result["result"]
                except:
                    logger.info(f'error: {async_tasks=}, {result=}, {_result=}')   
                    raise 
                # Зафиксировать время выполнения всех задач в блоке    
        elif isinstance(pgm, list):
            # Последовательно выполнить задачи
            __out_queue = None
            for stage in pgm:
                # Создать входную и выходную очередь для задач
                __in_queue = []
                # Скопировать первый раз входную очередь, далее результат предыдущего блока
                __in_queue.extend(in_queue if __out_queue is None else __out_queue)
                __out_queue = []
                # Выполнить пакет задач
                # logger.debug(f'- ++{stage}')
                _result = await self.execute(pgm = stage, in_queue = __in_queue, step = step + 2, backtrace = backtrace + ['-'])
                # logger.debug(f'- --{stage}')
                __out_queue = _result["queue"]
                stack_list = stack_list + [_result]
                result = _result["result"]
                # Прервать выполнение далее блоков программы
                # result == "exit"
                if 'exit' in result:
                    break
                # result == "when"
                if 'when' in result:
                    break
                # result == "break"
                if 'break' in result:
                    break
                # result == "continue"
                if 'continue' in result:
                    break
            result = list(set(result))    
            # result == "success"
            if ['success'] == result:    
                # Вернуть последнюю очередь, как результат, если все блоки были выполнены
                out_queue.extend(__out_queue)    
            # "when" флаг вычеркнуть из результата
            result = ['success' if 'when' == elem else elem for elem in result]    
        return self.context(result = list(set(result)), queue = out_queue, date_start = start_date, stack = stack_list if stack_list else stack_dict)

    def traces(self, in_obj, short = True):
        if isinstance(in_obj, list):
            return [self.traces(t, short) for t in in_obj]
        elif isinstance(in_obj, dict):
            if "stack" in in_obj:
                return self.traces(in_obj["stack"], short)
            else:
                res_dict = {}
                for key in in_obj:
                    res = self.traces(in_obj[key], short)
                    obj = in_obj[key]
                    params = []
                    if not short:
                        if "result" in obj:
                            params.append(f'r:[{",".join(obj["result"])}]')
                        if "param" in obj and obj["param"]:
                            params.append(f'p:{obj["param"]}')
                        if "queue" in obj and obj["queue"]:
                            params.append(f'q:{obj["queue"]}')
                    param = ", ".join(params)
                    if res is None:
                        if len(param) > 0:
                            res_dict[key] = param
                        else:    
                            res_dict[key] = None
                    else:
                        if len(param) > 0:
                            res_dict[f'{key} % {param}'] = res
                        else:    
                            res_dict[f'{key}'] = res
                return res_dict
                    
    def get_stack(self, short = True):   
        return self.traces(self.result, short)             

    def get_queue(self):   
        return self.result["queue"]

    def get_seconds(self):   
        delta = self.result["stop"] - self.result["start"]
        return delta.total_seconds()
    
    def get_data_keys(self):
        return mems.get_data_keys()

    async def aio_go(self, pgm:dict):
        in_queue = []
        self.result = await self.execute(pgm = pgm, in_queue = in_queue, backtrace = [])
        return self.result

    def __getattr__(self, attr):
        if attr == "queue":
            return self.get_queue()
        elif attr == "mem":
            return self.get_data_keys()
        elif attr == "seconds":
            return self.get_seconds()
        elif attr == "stack":
            return self.get_stack()
        elif attr == "opers":
            return {**(self.multiple_opers or {}), **(self.single_opers or {})}

    def save_short_stack(self, request = None):
        if request:
            file_name = f'{request.node.module.__name__} {request.node.name}'
        else:    
            file_name = f'test'
        save_object(self.get_stack(short = True), f"{file_name} - stack")

    def save_stack(self, request = None):
        if request:
            file_name = f'{request.node.module.__name__} {request.node.name}'
        else:    
            file_name = f'test'
        save_object(self.get_stack(short = False), f"{file_name} - stack")

def pgm(pgm_code:dict, pgm_libs:dict = None, in_classes:list = None, datas:dict = None, request = None):
    # Создать объект управления выполнением задач, загрузив исходный текст программы
    # save_object(value_dict = pgm_code, filename = "pgm_code", tp = "json")
    __pgm = Pgm()
    # Добавить обработчик
    if in_classes:
        for in_class in in_classes:
            __pgm.append_opers(in_class = in_class)
    # Разобрать код макросов    
    if pgm_libs:        
        if isinstance(pgm_libs, dict):
            if "macros" in pgm_libs:
                pgm_libs = pgm_libs["macros"]
            # Добавит макросы в задачу
            __pgm.append_macros(pgm_libs)
    # Преобразовать исходный код    
    if isinstance(pgm_code, dict):
        if "macros" in pgm_code:
            __pgm.append_macros(pgm_code["macros"])
        if "main" in pgm_code:
            pgm_code = pgm_code["main"]
    # Установить начальные переменные
    mems.data.clear()
    if datas:
        for key, value in datas.items():
            mems.set_data(key, value)
    # Запустить лавный модуль программы
    result = asyncio.get_event_loop().run_until_complete(__pgm.aio_go(pgm = pgm_code))
    # if request:
    #     file_name = f'{request.node.module.__name__} {request.node.name}'
    #     # zilog_utils.save_object(result, f"{file_name} - result")
    #     zilog_utils.save_object(pgm.get_stack(short = True), f"{file_name} - stack")
    # logger.info(__pgm.queue)
    return __pgm



import aiofiles
import yaml
import json
import os
import errno
import dataclasses
import datetime
import decimal
from datetime import date, datetime, timedelta
from typing import Any
# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

def get_attr(attr, elem, is_raise:bool = True):
    if attr:
        if attr and isinstance(elem, dict) and (attr in elem):
            return elem[attr]
        else:
            if is_raise:
                raise Exception(f'Не найден атрибут "{attr}" в элементе "{elem}"')    
            else:
                return None
    return elem    

def coalesce(plist: list):
    for elem in plist:
        if elem is not None:
            return elem
    return None    

def get_condition(test_value, param):
    """Проверить условие

    test_value:float:int:str - проверяемое значение
    param:dict - словарь заданных условий
    return bool - результат проверки
    """
    if isinstance(test_value, (float, int)):   
        where = True
        conditions_float = {
            "lt": lambda x, y: x <  y,
            "le": lambda x, y: x <= y,
            "eq": lambda x, y: x == y if y is not None else x is None,
            "ne": lambda x, y: x != y if y is not None else x is not None,
            "gt": lambda x, y: x >  y,
            "ge": lambda x, y: x >= y,
        }
        for key, function in conditions_float.items():
            if key in param.dict:
                where = where and function(test_value, param.get_float(key))
                logger.debug(f'{key=} {test_value=} {param.get_float(key)=} {where=}')
        return where
    elif isinstance(test_value, str):   
        where = True
        conditions_str = {
            "lt": lambda x, y: x <  y,
            "le": lambda x, y: x <= y,
            "eq": lambda x, y: x == y if y is not None else x is None,
            "ne": lambda x, y: x != y if y is not None else x is not None,
            "gt": lambda x, y: x >  y,
            "ge": lambda x, y: x >= y,
            # Входит строка
            "instr": lambda x, y: x in y,
            # Не входит строка
            "notinstr": lambda x, y: x not in y,
        }
        for key, function in conditions_str.items():
            if key in param.dict:
                where = where and function(test_value, param.get_string(key))
                logger.debug(f'{key=} {test_value=} {param.get_string(key)=} {where=}')
        return where
    return False



import asyncio
import re
import datetime
import aiofiles
from ..utils import aio_get_json, send_message, file_temp_path
from .util_opers import coalesce, get_condition

from ..param import Param, get_param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class OtherOpers:

    async def single_opers(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вывести словарь операций"""
        group = param.get_string()
        for oper in {**pgm.single_opers, **pgm.multiple_opers}.values():
            if group is None or group == oper["group"]:
                out_queue.append(oper)
        return ['success']

    async def single_popers(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вывести словарь операций"""
        if param.get_list("group"):
            group = param.get_list("group")
        elif param.get_string("group"):
            group = [param.get_string("group")]
        elif param.get_string():
            group = [param.get_string()]
        else:
            group = None    
        file_name = param.get_string("file") or (file_temp_path() + (f"/README.md {group}" if group else "/README.md"))
        for char in ['[', ']', "'"]:
            file_name = file_name.replace(char, "")
        if file_name:
            async with aiofiles.open(file_name, mode = 'w', encoding = "utf-8") as f:
                opers = []
                opers += pgm.single_opers.values()
                opers += pgm.multiple_opers.values()
                for group in sorted(list(set([oper["group"] for oper in opers if group is None or oper["group"] in group]))):
                    logger.info(f'## {group}')
                    await f.write(f'## {group}\n')
                    for oper_name in sorted([oper["name"] for oper in opers if oper["group"] == group]):
                        for oper in [oper for oper in opers if oper["name"] == oper_name]:
                            logger.info(f'### {oper["name"]} {"(multiple)" if oper["multiple"] else ""}')
                            await f.write(f'### {oper["name"]} {"(multiple)" if oper["multiple"] else ""}\n')
                            if oper["doc"]:
                                await f.write(f'```\n')
                                for line in oper["doc"].split('\n'):
                                    await f.write(f'{line.strip()}\n')
                                await f.write(f'```\n')
        return ['success']

    # async def single_desc(pgm, param, p_queue, in_queue = None, out_queue = None):
    #     out_queue.extend(in_queue)
    #     logger.info(f'desc: {param.string}')
    #     return ['success']

    async def _print_message(pgm, param, p_queue, in_queue, out_queue, tp):
        """Отправить сообщение об ошибке или успешное
        """
        out_queue.extend(in_queue)
        # Получить наименование тестируемого атрибута
        test_attr = param.get_string("test-attr")
        send_attr = param.get_string("send-attr")
        # Очередь ошибок
        errors = []
        for elem in in_queue:
            result = True
            if test_attr:
                pelem = Param(elem)
                test_value = coalesce([pelem.get_float(test_attr), pelem.get_string(test_attr)])
                result = get_condition(test_value, param)
            if result:    
                if send_attr:
                    if send_attr in elem:
                        errors.append(f'{elem[send_attr]}')    
                else:        
                    errors.append(f'{elem}')
        if len(errors):        
            name = param.get_string("name") or param.get_string()
            if tp:
                logger.info(f' {name}{": " if name else ""}{errors}')
            else:    
                logger.error(f' {name}{": " if name else ""}{errors}')
        return ['success']    

    async def single_info(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вывести очередь в info поток

        Входная очередь сообщение
        name:str = None or param:str = None - заголовок
        test-attr:str = None - тестируемый аттрибут
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        send-attr:str = None - отправляемый атрибут иначе печатаем весь элемент
        """
        return await OtherOpers._print_message(pgm, param, p_queue, in_queue, out_queue, 1)

    async def single_error(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вывести очередь в error поток

        Входная очередь сообщение
        name:str = None or param:str = None - заголовок
        test-attr:str = None - тестируемый аттрибут
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        send-attr:str = None - отправляемый атрибут иначе печатаем весь элемент
        """
        return await OtherOpers._print_message(pgm, param, p_queue, in_queue, out_queue, 0)

    async def single_sleep(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Асинхронно заснуть на заданное кол-во секунд
        
        param:float - кол-во секунд, если не задано, то 1 секунда
        """
        sleep_sec = param.get_float() or 1
        logger.info(f'sleep: {sleep_sec}')
        await asyncio.sleep(sleep_sec)
        out_queue.extend(in_queue)
        return ['success']

    async def multiple_msleep(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Асинхронно заснуть на заданное кол-во секунд. 
        Запустить несколько потоков, в зависимости от входной очереди.
        Для отладки асинхронности
        
        in_queue:list - очередь задержек
        """
        while len(in_queue):
            # Вычислить время засыпания
            sleep_sec = Param(in_queue.pop(0)).get_float()
            if sleep_sec:
                logger.info(f'msleep: {sleep_sec}')
                await asyncio.sleep(sleep_sec)
        return ['success']

    async def multiple_include(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Загрузить подпрограммы

        Входная очередь - имена json и yaml файлов
        """
        while len(in_queue):
            # Получить имя файла
            file_name = in_queue.pop(0)
            if isinstance(file_name, str):
                subroutines = await aio_get_json(file_name = file_name)
                for key, item in subroutines.items():
                    pgm.macros[key] = item 
                    logger.info(f'include: {key}')
        return ['success']    
   
    async def _send_message(pgm, param, p_queue, in_queue, out_queue, tp):
        """Отправить сообщение об ошибке или успешное
        """
        out_queue.extend(in_queue)
        # Получить наименование тестируемого атрибута
        test_attr = param.get_string("test-attr")
        send_attr = param.get_string("send-attr")
        # Очередь ошибок
        errors = []
        for elem in in_queue:
            result = True
            if test_attr:
                pelem = Param(elem)
                test_value = coalesce([pelem.get_float(test_attr), pelem.get_string(test_attr)])
                result = get_condition(test_value, param)
            if result:    
                if send_attr:
                    if send_attr in elem:
                        errors.append(f'{elem[send_attr]}')    
                else:        
                    errors.append(f'{elem}')
        if len(errors):        
            await send_message(message = "\n".join(errors), success = tp, service = param.get_string("name") or param.get_string() or  "PAQLang")        
        return ['success']    

    async def single_send_errors(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Отправить сообщение об ошибке
        
        Входная очередь сообщение
        name:str = None or param:str = None - сообщение в заголовке, по умолчанию PAQLang
        test-attr:str = None - тестируемый аттрибут
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        send-attr:str = None - отправляемый атрибут иначе печатаем весь элемент
        """
        return await OtherOpers._send_message(pgm = pgm, param = param, p_queue = p_queue, in_queue = in_queue, out_queue = out_queue, tp = 0)

    async def single_send_success(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Отправить успешное сообщение
        
        Входная очередь сообщение
        name:str = None or param:str = None - сообщение в заголовке, по умолчанию PAQLang
        test-attr:str = None - тестируемый аттрибут
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        send-attr:str = None - отправляемый атрибут иначе печатаем весь элемент
        """
        return await OtherOpers._send_message(pgm = pgm, param = param, p_queue = p_queue, in_queue = in_queue, out_queue = out_queue, tp = 1)



import asyncio
import datetime

from .util_opers import get_attr
from ..data import mems
from ..param import Param

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class DictOpers:

    def attr(l, elems, keys):
        if len(keys) == 0:
            return {}
        elif isinstance(elems, dict):
            result = {}
            for key in elems:
                if key in keys:
                    result[key] = elems[key]
            for key in elems:
                if isinstance(elems[key], list):
                    for elem in elems[key]:
                        x = DictOpers.attr(elem, keys)
                        if x:
                            result = {**result, **DictOpers.attr(elem, keys)}
            return result        

    def _plane(elem):
        if isinstance(elem, dict):
            result = {}
            ext = []
            for key in elem:
                el = elem[key]
                if isinstance(el, list):
                    for el1 in el:
                        ext.append(DictOpers._plane(el1))
                else:    
                    result[key] = el
            if len(ext) > 0:
                l = []
                for e in ext:
                    l.append({**result, **e})
                return l
            else:            
                return result

    async def single_plane(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Сделать плоскую таблицу. Каждый элемент рекурсивно раскрыть до элементов

        """
        for elem in in_queue:
            out_queue.extend(DictOpers._plane(elem))
        return ['success']

    async def single_attr(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить атрибут элемента

        param:str = None - если задан один атрибут, то вернуть его значение
        param:list = None - если задано несколько атрибутов, то вернуть словарь из этих атрибутов
        param:dict = None - вернуть набор заданных атрибутов, с возможностью замены на новый атрибут
        """
        # Задано два и более элемента
        if len(param.list) >= 2:
            # Перебрать очередь 
            for elem in in_queue:
                # Получить словарное представление очереди
                src_elem = Param(elem).dict
                result = {}
                # Перебрать все требуемы атрибуты
                for key in param.list:
                    # Атрибут найден, добавить его в словарь
                    if key in src_elem:
                        result[key] = src_elem[key]
                # Положить в результирующую очередь        
                if result:        
                    out_queue.append(result)
        #  Задан словарь            
        elif param.dict:
            # Перебрать элементы очереди
            for elem in in_queue:
                result = {}
                # Перебрать все требуемы атрибуты
                for key, result_key in param.dict.items():
                    # Если атрибут есть в очереди
                    if key in elem:
                        # Если задан атибут подмены
                        if result_key:
                            # Подменить ключ элемента
                            result[result_key] = elem[key]
                        else:    
                            # Добавить значение с атрибутом
                            result[key] = elem[key]
                # Положить в результирующую очередь        
                if result:        
                    out_queue.append(result)
        else:   
            # Получить наименование атрибута
            attr = param.get_string()
            if attr:
                # Перебрать очередь
                for elem in in_queue:
                    # Если атрибут входит в очередь, то вернуть в поток его значение
                    if isinstance(elem, dict) and attr in elem:
                        out_queue.append(elem[attr])

        return ['success']



import asyncio
import re
import datetime

from ..param import Param, get_param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class BoolOpers:

    async def single_bool_true(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Положить в очередь истина, любой не пустой массив
        """
        out_queue.extend([1])
        return ['success']
    
    async def single_bool_false(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Положить в очередь ложь. Пустой массив
        """
        out_queue.extend([])
        return ['success']

    async def single_not(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Инвертировать логическую операцию
        """
        if len(in_queue) == 0:
            out_queue.append(1)
        return ['success']



import asyncio
import datetime

from .util_opers import get_attr, get_condition
from ..param import Param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class MathOpers:

    async def single_filter(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Отобрать элементы из числовой очереди, удовлетворяющие условию. Не числовые элементы, отбрасываться из очереди

        attr:str - если указано, то анализировать атрибут из элемента очереди
        Условие параметры "lt", "le", "eq", "ne", "gt", "ge", "instr", "notinstr"
        Условия объединяются по "AND"
        """
        # Если задан параметр атрибут
        attr = param.get_string("attr")
        # Перебрать элементы очереди
        for elem in in_queue:
            # Получить проверяемое значение
            test_value = get_attr(attr = attr, elem = elem, is_raise = False) if attr else elem
            # Проверить условие
            result = get_condition(test_value, param)
            if result:
                out_queue.append(elem)
        return ['success']

    def __math_opers(param, in_queue, out_queue, msg, func):
        value = param.get_float()
        assert value is not None, msg
        for elem in in_queue:
            if isinstance(elem, (float, int)):   
                out_queue.append(func(elem, value))
        return ['success']

    async def single_add(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Добавить значение ко всем элементам очереди. Не числовые элементы, отбрасываться из очереди

        param:float - добавляемое значение
        """
        return MathOpers.__math_opers(param, in_queue, out_queue, "Не задано слогаемое", lambda x, y: x + y)

    async def single_sub(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вычесть значение из всех элементов очереди. Не числовые элементы, отбрасываться из очереди

        param:float - вычитаемое значение
        """
        return MathOpers.__math_opers(param, in_queue, out_queue, "Не задано вычитаемое значение", lambda x, y: x - y)

    async def single_mul(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Умножить на значение все элементы очереди. Не числовые элементы, отбрасываться из очереди

        param:float - множитель
        """
        return MathOpers.__math_opers(param, in_queue, out_queue, "Не задан множитель", lambda x, y: x * y)

    async def single_div(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Разделить на значение все элементы очереди. Не числовые элементы, отбрасываться из очереди

        param:float - делитель
        """
        assert param.get_float() != 0, "Делитель равен нулю"
        return MathOpers.__math_opers(param, in_queue, out_queue, "Не задан делитель", lambda x, y: x / y)

    def __inc_dec(param, p_queue, out_queue, sign):
        # Получить шаг приращения, атрибут, параметр в функции или 1
        step = param.get_float("step") or 1
        # Получить имя переменной
        mem_name = param.get_string("mem") or param.get_string()
        if mem_name:
            # Если переменная существует, то взять из нее значение, иначе 0
            result = Param(mems.get_data(mem_name)).float if mem_name in mems.data else 0
        else:
            # Взять значение из очереди
            result = p_queue.float    
        # Операция приращения    
        result += sign * step  
        # Записать значение в переменную
        if mem_name:
            mems.set_data(mem_name, [result])    
        out_queue.extend([result])
        return ['success']

    async def single_inc(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Добавить к очереди, числовое значение

        step:float или param:float или 1 - добавляемое значение
        mem:str - если указана переменная, то добавить значение к переменной, результат также положить в очередь
        """
        return MathOpers.__inc_dec(param = param, p_queue = p_queue, out_queue = out_queue, sign = 1)

    async def single_dec(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Уменьшить значение в очереди

        step:float или param:float или 1 - уменьшаемое значение
        mem:str - если указана переменная, то уменьшить значение переменной, результат также положить в очередь
        """
        return MathOpers.__inc_dec(param = param, p_queue = p_queue, out_queue = out_queue, sign = -1)



import asyncio
import re
import datetime
import os
from ..utils import aio_reads, get_json
from .util_opers import get_attr

from ..param import Param, get_param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class IoOpers:

    async def single_walk(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить список файлов в директории
        
        path:str или param:str - корневой маршрут поиска файлов
        regex:str - шаблон отбора файлов, если не задан, то все файлы
        """
        regex = param.get_string("regex")
        while len(in_queue):
            # Получить файл
            path = in_queue.pop(0)
            if isinstance(path, str):
                for root, dirs, files in os.walk(path):
                    for file_name in [os.path.join(root, name) for name in files]:
                        if regex and not re.match(regex, file_name):
                                continue
                        out_queue.append(file_name)
        return ['success']

    async def multiple_freads(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Прочитать файл и положить содержимое в выходную очередь {fname, text}

        encoding:str, если не задано, то param:str - по умолчанию utf-8
        кодировка для windows файлов cp1251
        """
        encoding = param.get_string("encoding") or param.get_string()
        while len(in_queue):
            # Получить файл
            file_name = in_queue.pop(0)
            if isinstance(file_name, str):
                out_queue.append({
                    "path": file_name, 
                    "text": await aio_reads(file_name = file_name, encoding = encoding)})
        return ['success']

    async def single_to_json(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Распарсить JSON или YAML строковый элемент и положить в очередь

        Входная строка преобразуется в JSON, несколько массивов склеиваются
        attr:str = None или param:str = None - задает имя атрибута текста JSON или YAML
        """
        # Если задан параметр атрибут
        attr = param.get_string("attr") or param.get_string()
        while len(in_queue):
            # Получить элемент из очереди, с начала списка
            text = in_queue.pop(0)
            js_text = get_attr(attr = attr, elem = text, is_raise = False) if attr else text
            if js_text:
                js = get_json(text = js_text)
                if isinstance(js, list):
                    for elem in js:
                        if elem:
                            out_queue.append(elem)
                else:
                    out_queue.append(js)
        return ['success']



import asyncio
import re
import datetime

from ..utils import save_object
from ..param import Param, get_param
from ..data import mems

# Стандартное логирование
import logging
# Установить текущи логгер
logger = logging.getLogger(__name__)

class QueueOpers:

    async def single_print(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Распечатать текущую очередь данных, в loggin.info

        in_queue - входная очередь
        ps: очередь не изменяется
        """
        name = param.get_string()
        for elem in in_queue:
            if name:
                logger.info(f'print({name}): {elem}')
            else:    
                logger.info(f'print: {elem}')
        out_queue.extend(in_queue)
        return ['success']

    async def single_in(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Положить данные в очередь как список
        
        Результат в выходную очередь out_queue
        """
        out_queue.extend(param.list)
        # logger.info(f'in: {len(param.list)}')
        return ['success']

    async def single_out(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Содержимое очереди записать в переменную памяти

        in_queue - входная очередь
        param:str - наименование переменной памяти
        ps: очередь не изменяется
        """
        out_queue.extend(in_queue)
        mem_name = param.get_string()
        assert mem_name is not None, "Не указано наименование переменной памяти"
        logger.info(f'out: {mem_name}')
        mems.set_data(mem_name, in_queue)
        return ['success']
    
    async def single_sort(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Отсортировать список
        
        in_queue - входная очередь
        param:str - направление сортировки "desc", "asc" - по умолчанию
        """
        if param.get_string() and param.get_string() == "desc":
            out_queue.extend(sorted(in_queue, reverse = True))
        else:    
            out_queue.extend(sorted(in_queue))
        return ['success']

    async def single_distinct(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Получить уникальные элементы очереди и отсортировать по возрастанию
        
        in_queue - входная очередь
        """
        out_queue.extend(sorted(list(set(in_queue))))
        return ['success']

    async def single_len(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вернуть размер входной очереди
        
        in_queue - входная очередь
        """
        out_queue.append(len(in_queue))
        return ['success']

    async def single_push(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Добавить элементы в конец текущей очереди

        in_queue - входная очередь
        param - добавляемое значение в очередь
        """
        out_queue.extend(in_queue)
        out_queue.extend(param.list)
        return ['success']

    async def single_first(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вернуть первый элемент очереди
        
        in_queue - входная очередь
        param:str - если задано значение, то взять очередь из переменной
        """
        mem_name = param.get_string()
        if mem_name:
            __in_queue = mems.get_data(mem_name)
        else:    
            __in_queue = in_queue
        if len(__in_queue) > 0:
            out_queue.append(__in_queue.pop(0))
        return ['success']

    async def single_last(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вернуть последний элемент очереди

        in_queue - входная очередь
        param:str - если задано значение, то взять очередь из переменной
        """
        mem_name = param.get_string()
        if mem_name:
            __in_queue = mems.get_data(mem_name)
        else:    
            __in_queue = in_queue
        if len(__in_queue) > 0:
            out_queue.append(__in_queue.pop())
        return ['success']

    async def single_pop(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вытолкнуть последний элемент из очереди

        in_queue - входная очередь
        param:str - если задано значение, то взять очередь из переменной
        """
        mem_name = param.get_string()
        if mem_name:
            __in_queue = mems.get_data(mem_name)
        else:    
            __in_queue = in_queue
        if len(__in_queue) > 0:
            out_queue.append(__in_queue.pop())
        if mem_name:
            mems.set_data(mem_name, __in_queue)
        return ['success']

    async def single_save(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Сохранить очередь на диске

        in_queue - входная очередь
        param:str - наименование файла данных
        """
        out_queue.extend(in_queue)
        save_name = param.get_string()
        assert save_name is not None, "Не указано имя файла результата"
        save_object(value_dict = in_queue, filename = save_name)
        return ['success']

    async def single_minus(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вычесть очередь из очереди

        in_queue - входная очередь
        param:list - вычитаемая очередь
        """
        # Скопировать очередь
        __in_queue = in_queue.copy()
        # Пробежать вычитаемую очередь
        for x in param.list:
            # Если есть элемент во входной очереди
            while x in __in_queue:
                # Убрать элемент из исходной очереди
                __in_queue.remove(x)
        # Положить элемент во входную очередь        
        out_queue.extend(__in_queue)
        return ['success']

    async def single_intersect(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Вернуть только пересекаемые элементы очереди

        in_queue - первая очередь
        param:list - вторая очередь
        """
        out_queue.extend(list(set(in_queue) & set(param.list)))
        return ['success']

    async def single_union_all(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Объединить очереди

        in_queue - входная очередь
        param - добавляемое значение в очередь
        """
        out_queue.extend(in_queue)
        out_queue.extend(param.list)
        return ['success']

    async def single_union(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Объединить очереди, убрать дубликаты

        in_queue - входная очередь
        param - добавляемое значение в очередь
        """
        # Скопировать очередь
        __in_queue = in_queue.copy()
        # Добавить элементы в очередь
        __in_queue.extend(param.list)
        # Добавить элементы в очередь, убрав дубликаты
        out_queue.extend(list(set(__in_queue)))
        return ['success']

    async def single_expand(pgm, param, p_queue, in_queue = None, out_queue = None):
        """Расширить подмассивы

        in_queue - входная очередь
        param - добавляемое значение в очередь
        """
        # Скопировать очередь
        for item in in_queue:
            if isinstance(item, list):
                out_queue.extend(item)
            else:    
                out_queue.append(item)
        return ['success']
